from typing import Any, Dict, List, Optional, Tuple, Type, Union, overload

import jax
import numpy as np
from jax.tree_util import register_pytree_node

from ciclo.timetracking import Elapsed
from ciclo.types import CluMetric, LogPath

LogsLike = Dict[str, Dict[str, Any]]
Collection = str
Entry = str


class Entries(Dict[Entry, Any]):
    def __getattr__(self, entry: Entry) -> Any:
        if entry in self:
            return self[entry]
        else:
            raise AttributeError(f"Entry '{entry}' not found in logs.")


class Logs(Dict[Collection, Entries]):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        # copy values
        for k, v in self.items():
            self[k] = Entries(v)

    @property
    def updates(self) -> Optional[LogsLike]:
        # raise error if accessed
        raise AttributeError("updates is a write-only attribute")

    @updates.setter
    def updates(self, updates: Optional[LogsLike]) -> None:
        if updates is not None:
            self.merge(updates)

    def __getattr__(self, collection: Collection) -> Entries:
        if collection in self:
            return self[collection]
        else:
            raise AttributeError(f"Collection '{collection}' not found in logs.")

    # ----------------------------------
    # logger behavior
    # ----------------------------------
    def add_entry(self, collection: Collection, entry: Entry, value: Any) -> "Logs":
        if collection not in self:
            self[collection] = Entries()
        self[collection][entry] = value
        return self

    def add_entries(self, collection: Collection, **values: Any) -> "Logs":
        for name, value in values.items():
            self.add_entry(collection, name, value)

        return self

    def add_metric(self, entry: Entry, value: Any) -> "Logs":
        if isinstance(value, CluMetric):
            raise ValueError(
                f"Metric '{entry}' is a clu Metric which is "
                "stateful, use 'add_stateful_metric' instead"
            )
        return self.add_entry("metrics", entry, value)

    def add_metrics(self, **metrics: Any):
        return self.add_entries("metrics", **metrics)

    def add_stateful_metric(self, name: Entry, value: Any) -> "Logs":
        return self.add_entry("stateful_metrics", name, value)

    def add_stateful_metrics(self, **metrics: Any) -> "Logs":
        return self.add_entries("stateful_metrics", **metrics)

    def add_loss(self, name: Entry, value: Any, *, add_metric: bool = False) -> "Logs":
        self.add_entry("losses", name, value)
        if add_metric:
            self.add_metric(name, value)
        return self

    def add_losses(
        self,
        *,
        add_metrics: bool = False,
        **losses: Any,
    ) -> "Logs":
        self.add_entries("losses", **losses)
        if add_metrics:
            self.add_metrics(**losses)
        return self

    def add_output(self, name: Entry, value: Any) -> "Logs":
        return self.add_entry("outputs", name, value)

    def add_outputs(self, **outputs: Any) -> "Logs":
        return self.add_entries("outputs", **outputs)

    # ----------------------------------
    # history behavior
    # ----------------------------------

    def entry_value(self, name: Entry) -> Any:
        path = self.entry_path(name)
        if path is None:
            raise KeyError(f"Key {name} not found in logs.")
        collection, name = path
        return self[collection][name]

    def entry_path(self, name: Entry) -> Optional[LogPath]:
        path = name.split(".")

        if len(path) == 1:
            name = path[0]
            collection = self.entry_collection(name)
            if collection is None:
                return None
        elif len(path) == 2:
            collection, name = path
        else:
            raise ValueError(f"Got more than 2 levels of nesting in key '{name}'")

        return collection, name

    def entry_collection(self, name: Entry) -> Optional[str]:
        collections = [col for col in self if name in self[col]]

        if len(collections) == 0:
            return None
        elif len(collections) == 1:
            return collections[0]
        else:
            raise ValueError(
                f"Found multiple collections for name '{name}' : {collections}. "
                "Use `collection.name` syntax."
            )

    def merge(self, collection_updates: LogsLike):
        for collection, updates in collection_updates.items():
            if collection in self:
                self[collection].update(updates)
            else:
                self[collection] = Entries(updates)  # create copy


def _mapping_tree_flatten(tree: Dict[Any, Any]):
    return (dict(tree),), None


def _mapping_tree_unflatten(cls: Type[Dict[Any, Any]], children: Tuple[Dict[str, Any]]):
    tree = cls(children[0])
    return tree


register_pytree_node(
    Logs,
    _mapping_tree_flatten,
    lambda _, children: _mapping_tree_unflatten(Logs, children),
)
register_pytree_node(
    Entries,
    _mapping_tree_flatten,
    lambda _, children: _mapping_tree_unflatten(Entries, children),
)

# ----------------------------------
# history
# ----------------------------------


class History(List[Logs]):
    @overload
    def collect(self, key: str) -> List[Any]:
        ...

    @overload
    def collect(self, *keys: str) -> Tuple[List[Any], ...]:
        ...

    def collect(self, key: str, *keys: str) -> Union[List[Any], Tuple[List[Any], ...]]:
        keys = (key,) + keys
        outputs = tuple([] for _ in keys)
        for logs in self:
            paths = [logs.entry_path(key) for key in keys]
            if all(path is not None for path in paths):
                for i, path in enumerate(paths):
                    assert path is not None
                    collection, key = path
                    outputs[i].append(logs[collection][key])

        return outputs if len(keys) > 1 else outputs[0]

    def commit(self, elapsed: Elapsed, logs: LogsLike):
        # convert JAX arrays to numpy arrays to free memory
        logs = jax.tree_map(
            lambda x: np.asarray(x) if isinstance(x, jax.Array) else x, logs
        )
        logs["elapsed"] = elapsed.to_dict()

        self.append(Logs(logs))



import abc
import dataclasses
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Protocol, Tuple, Union, runtime_checkable

from ciclo.timetracking import Elapsed, Period
from ciclo.types import A, Schedule

ScheduleLike = Schedule


@runtime_checkable
class Copyable(Protocol):
    def copy(self: A) -> A:
        ...


# ---------------------------------------
# base
# ---------------------------------------


class ScheduleBase(abc.ABC, Copyable):
    @abc.abstractmethod
    def __call__(self, elapsed: Elapsed) -> bool:
        pass

    @abc.abstractmethod
    def copy(self: A) -> A:
        ...

    def __and__(self, other: Schedule) -> "And":
        return And.create(self, other)

    def __or__(self, other: Schedule) -> "Or":
        return Or.create(self, other)

    def __not__(self) -> "Not":
        return Not.create(self)


# ---------------------------------------
# logical
# ---------------------------------------


@dataclass(unsafe_hash=True)
class And(ScheduleBase):
    schedules: Tuple[Schedule, ...]

    @classmethod
    def create(cls, *schedules: ScheduleLike) -> "And":
        return cls(schedules=tuple(map(to_schedule, schedules)))

    def __call__(self, elapsed: Elapsed) -> bool:
        return all(schedule(elapsed) for schedule in self.schedules)

    def copy(self) -> "And":
        return dataclasses.replace(self)


@dataclass(unsafe_hash=True)
class Or(ScheduleBase):
    schedules: Tuple[Schedule, ...]

    @classmethod
    def create(cls, *schedules: ScheduleLike) -> "Or":
        return Or(schedules=tuple(map(to_schedule, schedules)))

    def __call__(self, elapsed: Elapsed) -> bool:
        return any(schedule(elapsed) for schedule in self.schedules)

    def copy(self) -> "Or":
        return dataclasses.replace(self)


@dataclass(unsafe_hash=True)
class Not(ScheduleBase):
    schedule: Schedule

    @classmethod
    def create(cls, schedule: Schedule) -> "Not":
        return cls(schedule=to_schedule(schedule))

    def __call__(self, elapsed: Elapsed) -> bool:
        return not self.schedule(elapsed)

    def copy(self) -> "Not":
        return dataclasses.replace(self)


@dataclass(unsafe_hash=True)
class Constant(ScheduleBase):
    value: bool

    @classmethod
    def create(cls, value: bool) -> "Constant":
        return cls(value=value)

    def __call__(self, elapsed: Elapsed) -> bool:
        return self.value

    def copy(self) -> "Constant":
        return dataclasses.replace(self)


@dataclass(unsafe_hash=True)
class Lambda(ScheduleBase):
    func: Schedule

    @classmethod
    def create(cls, func: Schedule) -> "Lambda":
        return cls(func=func)

    def __call__(self, elapsed: Elapsed) -> bool:
        return self.func(elapsed)

    def copy(self) -> "Lambda":
        return dataclasses.replace(self)


def always() -> Constant:
    return Constant.create(True)


def never() -> Constant:
    return Constant.create(False)


# ---------------------------------------
# every
# ---------------------------------------


def every(
    steps: Union[int, None] = None,
    *,
    samples: Union[int, None] = None,
    time: Union[timedelta, float, int, None] = None,
) -> "Every":
    return Every(
        period=Period.create(steps=steps, samples=samples, time=time),
        last_samples=0,
        last_time=datetime.now().timestamp(),
    )


@dataclass(unsafe_hash=True)
class Every(ScheduleBase):
    period: Period
    last_samples: int
    last_time: float

    def __call__(self, elapsed: Elapsed) -> bool:
        if self.period.steps is not None:
            steps = elapsed.steps
            return steps >= 0 and steps % self.period.steps == 0

        if self.period.samples is not None:
            if elapsed.samples - self.last_samples >= self.period.samples:
                self.last_samples = elapsed.samples
                return True

        if self.period.time is not None:
            if elapsed.date - self.last_time >= self.period.time:
                self.last_time = elapsed.date
                return True

        return False

    def copy(self) -> "Every":
        return dataclasses.replace(self)


# ---------------------------------------
# after
# ---------------------------------------


def after(
    steps: Union[int, None] = None,
    *,
    samples: Union[int, None] = None,
    time: Union[timedelta, float, int, None] = None,
) -> "After":
    return After(
        period=Period.create(steps=steps, samples=samples, time=time),
    )


@dataclass(unsafe_hash=True)
class After(ScheduleBase):
    period: Period

    def __call__(self, elapsed: Elapsed) -> bool:
        return elapsed >= self.period

    def copy(self) -> "After":
        return dataclasses.replace(self)

    def then(self, schedule: Schedule) -> "Piecewise":
        return piecewise(never(), {self.period: schedule})

    def __rshift__(self, schedule: Schedule) -> "Piecewise":
        return self.then(schedule)

    def every(
        self,
        steps: Optional[int] = None,
        *,
        samples: Optional[int] = None,
        time: Optional[Union[timedelta, float, int]] = None,
    ) -> "Piecewise":
        return self.then(every(steps=steps, samples=samples, time=time))


# ---------------------------------------
# piecewise
# ---------------------------------------


def piecewise(
    schedule: Schedule,
    period_schedules: Dict[Period, Schedule],
) -> "Piecewise":
    return Piecewise(
        schedule=to_schedule(schedule),
        period_schedules=tuple(
            (p, to_schedule(s)) for p, s in period_schedules.items()
        ),
    )


@dataclass(unsafe_hash=True)
class Piecewise(ScheduleBase):
    schedule: Schedule
    period_schedules: Tuple[Tuple[Period, Schedule], ...]
    reference: Optional[Elapsed] = None
    schedule_index: int = -1

    @property
    def current_schedule(self) -> Schedule:
        if self.schedule_index == -1:
            return self.schedule
        else:
            return self.period_schedules[self.schedule_index][1]

    @property
    def current_period(self) -> Period:
        return self.period_schedules[self.schedule_index + 1][0]

    def __call__(self, elapsed: Elapsed) -> bool:
        if self.reference is None:
            self.reference = elapsed

        if self.schedule_index < len(self.period_schedules) - 1:
            if elapsed >= self.current_period:
                self.schedule_index += 1
                self.reference = elapsed

        elapsed = elapsed - self.reference
        return self.current_schedule(elapsed)

    def copy(self) -> "Piecewise":
        return dataclasses.replace(self)


# ---------------------------------------
# utils
# ---------------------------------------


def to_schedule(value: Any) -> Schedule:
    value = maybe_copy(value)

    if isinstance(value, ScheduleBase):
        return value
    elif callable(value):
        return Lambda.create(value)
    # elif isinstance(value, bool):
    #     return Constant.create(value)
    # elif isinstance(value, int):
    #     return every(steps=value)
    else:
        raise ValueError(
            f"Invalid schedule, must be a Schedule, callable, bool or int, got: {value}"
        )


def maybe_copy(value: A) -> A:
    if isinstance(value, Copyable):
        return value.copy()
    else:
        return value



import dataclasses
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union

from flax import struct

from ciclo.types import B


class Elapsed(struct.PyTreeNode):
    steps: int
    samples: int
    date: float
    date_start: float = struct.field(pytree_node=True, repr=False)

    @property
    def time(self) -> float:
        return self.date - self.date_start

    @classmethod
    def create(cls, steps: int = 0, samples: int = 0) -> "Elapsed":
        now = datetime.now().timestamp()
        return cls(steps=steps, samples=samples, date_start=now, date=now)

    def update(self, batch_size: int) -> "Elapsed":
        return self.replace(
            steps=self.steps + 1,
            samples=self.samples + batch_size,
            date=datetime.now().timestamp(),
        )

    def update_time(self) -> "Elapsed":
        return self.replace(date=datetime.now().timestamp())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "steps": self.steps,
            "samples": self.samples,
            "date": self.date,
            "date_start": self.date_start,
            "time": self.time,
        }

    def __sub__(self, other: "Elapsed") -> "Elapsed":
        return Elapsed(
            steps=self.steps - other.steps,
            samples=self.samples - other.samples,
            date=self.date,
            date_start=other.date_start,
        )

    def _compare(
        self, other: "Period", predicate: Callable[[float, float], bool]
    ) -> bool:
        if other.steps is not None and predicate(self.steps, other.steps):
            return True
        if other.samples is not None and predicate(self.samples, other.samples):
            return True
        if other.time is not None and predicate(self.time, other.time):
            return True
        if other.date is not None and predicate(self.date, other.date):
            return True
        return False

    def __ge__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a >= b)

    def __gt__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a > b)

    def __le__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a <= b)

    def __lt__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a < b)

    def __eq__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a == b)

    def __ne__(self, other: "Period") -> bool:
        return self._compare(other, lambda a, b: a != b)


@dataclasses.dataclass(frozen=True)
class Period:
    steps: Union[int, None]
    samples: Union[int, None]
    time: Union[float, int, None]
    date: Union[float, None]

    @classmethod
    def create(
        cls,
        steps: Union[int, None] = None,
        samples: Union[int, None] = None,
        time: Union[timedelta, float, int, None] = None,
        date: Union[datetime, float, None] = None,
    ):
        if all(x is None for x in [steps, samples, time, date]):
            raise ValueError("At least one duration parameter must be specified.")

        return cls(
            steps=steps,
            samples=samples,
            time=time.total_seconds() if isinstance(time, timedelta) else time,
            date=date.timestamp() if isinstance(date, datetime) else date,
        )

    def __repr__(self) -> str:
        params_repr = ", ".join(
            f"{k}={v}" for k, v in self.__dict__.items() if v is not None
        )
        return f"Period({params_repr})"


PeriodLike = Union[Period, int]


def to_period(period: PeriodLike) -> Period:
    if isinstance(period, int):
        period = Period.create(steps=period)
    return period


def elapse(
    dataset: Iterable[B],
    initial: Optional[Elapsed] = None,
    stop: Optional[PeriodLike] = None,
    batch_size_fn: Optional[Callable[[List[Tuple[int, ...]]], int]] = None,
) -> Iterable[Tuple[Elapsed, B]]:
    from ciclo.utils import get_batch_size, max_first_axis

    if batch_size_fn is None:
        batch_size_fn = max_first_axis

    if stop is not None:
        stop = to_period(stop)

    elapsed = initial or Elapsed.create()
    for batch in dataset:
        yield elapsed, batch
        batch_size = get_batch_size(batch, batch_size_fn=batch_size_fn)
        elapsed = elapsed.update(batch_size)

        if stop and elapsed >= stop:
            break



import functools
import importlib.util
import os
from dataclasses import dataclass, replace
from datetime import datetime
from enum import Enum, auto
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Union,
    overload,
)

from pkbar import Kbar
from tqdm import tqdm

from ciclo.logging import Collection, Entry, History, Logs
from ciclo.loops.loop import (
    CallbackOutput,
    LoopCallbackBase,
    LoopOutput,
    LoopState,
    register_adapter,
)
from ciclo.schedules import every
from ciclo.timetracking import Elapsed, Period
from ciclo.types import Batch, S
from ciclo.utils import get_batch_size, is_scalar

AggregationFn = Callable[[List[Any]], Any]
InnerLoopAggregation = Union[
    Literal["last", "mean", "sum", "min", "max", "first"], AggregationFn
]


def unavailable_dependency(msg: str) -> Any:
    class DependencyNotAvailable(LoopCallbackBase[S]):
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            raise RuntimeError(msg)

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            raise RuntimeError(msg)

    return DependencyNotAvailable


class OptimizationMode(str, Enum):
    min = auto()
    max = auto()


def _transpose_history(
    log_history: History,
) -> Mapping[Collection, Mapping[Entry, List[Any]]]:
    """Convert a list of (nested) log dictionaries into a (nested) dictionary of lists."""
    result = {}
    for log_dict in log_history:
        for collection, entries in log_dict.items():
            if collection not in result:
                result[collection] = {}
            for entry, value in entries.items():
                if entry not in result[collection]:
                    result[collection][entry] = []
                result[collection][entry].append(value)
    return result


class inner_loop(LoopCallbackBase[S]):
    @overload
    def __init__(
        self,
        name_or_loop_fn: str,
        maybe_loop_fn: Callable[[S], LoopOutput[S]],
        *,
        output_state: bool = False,
    ):
        ...

    @overload
    def __init__(
        self,
        name_or_loop_fn: Callable[[S], LoopOutput[S]],
        *,
        output_state: bool = False,
    ):
        ...

    def __init__(
        self,
        name_or_loop_fn: Union[str, Callable[[S], LoopOutput[S]]],
        maybe_loop_fn: Optional[Callable[[S], LoopOutput[S]]] = None,
        *,
        output_state: bool = False,
        aggregation: Union[
            InnerLoopAggregation, Mapping[Collection, InnerLoopAggregation]
        ] = "last",
    ):
        if isinstance(name_or_loop_fn, str):
            assert maybe_loop_fn is not None
            self.name = name_or_loop_fn
            self.loop_fn = maybe_loop_fn
        else:
            assert maybe_loop_fn is None
            self.name = None
            self.loop_fn = name_or_loop_fn
        self.output_state = output_state
        self.aggregation = aggregation

    def __call__(self, state: S) -> Tuple[Logs, S]:
        inner_state, log_history, _ = self.loop_fn(state)
        logs = _transpose_history(log_history)
        logs = Logs(
            {
                collection: {
                    entry + f"_{self.name}"
                    if self.name
                    else entry: self.__get_aggregation_fn(collection)(values)
                    for entry, values in entries.items()
                }
                for collection, entries in logs.items()
                if collection != "elapsed"
            }
        )
        return logs, (inner_state if self.output_state else state)

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        return self(loop_state.state)

    def __get_aggregation_fn(self, collection: Collection) -> AggregationFn:
        if isinstance(self.aggregation, Mapping):
            aggregation = self.aggregation.get(collection, "last")
            error_message = f"The aggregation ({aggregation}) for collection {collection} must be a str or Callable."
        else:
            aggregation = self.aggregation
            error_message = (
                f"The aggregation ({aggregation}) must be a str or Callable."
            )

        if not (isinstance(aggregation, str) or isinstance(aggregation, Callable)):
            raise ValueError(error_message)

        if aggregation == "last":
            return lambda x: x[-1]
        elif aggregation == "mean":
            return lambda x: sum(x) / len(x)
        elif aggregation == "sum":
            return sum
        elif aggregation == "min":
            return min
        elif aggregation == "max":
            return max
        elif aggregation == "first":
            return lambda x: x[0]
        elif isinstance(aggregation, Callable):
            return aggregation
        else:
            raise ValueError(f"Invalid aggregation: {aggregation}")


if importlib.util.find_spec("tensorflow") is not None:
    from flax.training import checkpoints as flax_checkpoints

    class checkpoint(LoopCallbackBase[S]):
        def __init__(
            self,
            ckpt_dir: Union[str, os.PathLike],
            prefix: str = "checkpoint_",
            keep: int = 1,
            overwrite: bool = False,
            keep_every_n_steps: Optional[int] = None,
            async_manager: Optional[flax_checkpoints.AsyncManager] = None,
            monitor: Optional[str] = None,
            mode: Union[str, OptimizationMode] = "min",
        ):
            if isinstance(mode, str):
                mode = OptimizationMode[mode]

            if mode not in OptimizationMode:
                raise ValueError(
                    f"Invalid mode: {mode}, expected one of {list(OptimizationMode)}"
                )
            else:
                self.mode = mode

            self.ckpt_dir = ckpt_dir
            self.prefix = prefix
            self.keep = keep
            self.overwrite = overwrite
            self.keep_every_n_steps = keep_every_n_steps
            self.async_manager = async_manager
            self.monitor = monitor
            self.minimize = self.mode == OptimizationMode.min
            self._best: Optional[float] = None

        def __call__(
            self, elapsed: Elapsed, state: S, logs: Optional[Logs] = None
        ) -> None:
            save_checkpoint = True
            step_or_metric = elapsed.steps
            overwrite = self.overwrite

            if self.monitor is not None:
                if logs is None:
                    raise ValueError(
                        "checkpoint callback requires logs to monitor a metric"
                    )
                if not isinstance(logs, Logs):
                    logs = Logs(logs)

                try:
                    value = logs.entry_value(self.monitor)
                except KeyError:
                    raise ValueError(
                        f"Monitored value '{self.monitor}' not found in logs"
                    )

                if (
                    self._best is None
                    or (self.minimize and value < self._best)
                    or (not self.minimize and value > self._best)
                ):
                    self._best = value
                    step_or_metric = (
                        value if self.mode == OptimizationMode.max else -value
                    )
                else:
                    save_checkpoint = False

            if save_checkpoint:
                flax_checkpoints.save_checkpoint(
                    ckpt_dir=self.ckpt_dir,
                    target=state,
                    step=step_or_metric,
                    prefix=self.prefix,
                    keep=self.keep,
                    overwrite=overwrite,
                    keep_every_n_steps=self.keep_every_n_steps,
                    async_manager=self.async_manager,
                )

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self(loop_state.elapsed, loop_state.state, loop_state.accumulated_logs)
            return Logs(), loop_state.state

        def on_epoch_end(
            self, state, batch, elapsed, loop_state: LoopState[S]
        ) -> CallbackOutput[S]:
            return self.__loop_callback__(loop_state)

else:
    checkpoint = unavailable_dependency(
        "'tensorflow' package is not available, please install it to use the 'checkpoint' callback"
    )


class early_stopping(LoopCallbackBase[S]):
    def __init__(
        self,
        monitor: str,
        patience: Union[int, Period],
        min_delta: float = 0,
        mode: Union[str, OptimizationMode] = "min",
        baseline: Optional[float] = None,
        restore_best_weights: bool = False,
    ):
        if isinstance(mode, str):
            mode = OptimizationMode[mode]

        if mode not in OptimizationMode:
            raise ValueError(
                f"Invalid mode: {mode}, expected one of {list(OptimizationMode)}"
            )
        else:
            self.mode = mode

        self.monitor = monitor
        self.patience = (
            patience if isinstance(patience, Period) else Period.create(patience)
        )
        self.min_delta = min_delta
        self.mode = mode
        self.baseline = baseline
        self.restore_best_weights = restore_best_weights
        self.minimize = self.mode == OptimizationMode.min
        self._best = baseline
        self._best_state = None
        self._elapsed_start: Optional[Elapsed] = None

    def __call__(self, elapsed: Elapsed, state: S, logs: Logs) -> Tuple[bool, S]:
        stop_iteration = False

        if self._elapsed_start is None:
            self._elapsed_start = elapsed

        if not isinstance(logs, Logs):
            logs = Logs(logs)

        try:
            value = logs.entry_value(self.monitor)
        except KeyError:
            raise ValueError(f"Monitored value '{self.monitor}' not found in logs")

        if (
            self._best is None
            or (self.minimize and value < self._best)
            or (not self.minimize and value > self._best)
        ):
            self._best = value
            self._best_state = state
            self._elapsed_start = elapsed

        if elapsed - self._elapsed_start >= self.patience:
            if self.restore_best_weights and self._best_state is not None:
                state = self._best_state
            stop_iteration = True

        return stop_iteration, state

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        stop_iteration, state = self(
            loop_state.elapsed, loop_state.state, loop_state.accumulated_logs
        )
        loop_state.stop_iteration = stop_iteration
        return Logs(), state


class tqdm_bar(LoopCallbackBase[S]):
    def __init__(
        self,
        total: Union[Period, int, None] = None,
        desc=None,
        leave=True,
        file=None,
        ncols=None,
        mininterval=0.1,
        maxinterval=10.0,
        miniters=None,
        ascii=None,
        disable=False,
        unit_scale=False,
        dynamic_ncols=False,
        smoothing=0.3,
        bar_format=None,
        initial=0,
        position=None,
        postfix=None,
        unit_divisor=1000,
        write_bytes=None,
        lock_args=None,
        nrows=None,
        colour=None,
        delay=0,
        gui=False,
        **kwargs,
    ):
        if isinstance(total, int):
            total = Period.create(steps=total)

        if total is not None:
            if total.steps is not None:
                bar_total = total.steps
                unit = "steps"
            elif total.samples is not None:
                bar_total = total.samples
                unit = "samples"
            elif total.time is not None:
                bar_total = total.time
                unit = "s"
                unit_scale = True
            elif total.date is not None:
                total = replace(total, time=total.date - datetime.now().timestamp())
                bar_total = total.time
                unit = "s"
                unit_scale = True
            else:
                raise ValueError("Invalid total")
        else:
            bar_total = None
            unit = "it"

        self.total = total
        self.prev_step: Optional[int] = None
        self.prev_samples: Optional[int] = None
        self.prev_time: Optional[float] = None
        self.bar_total = bar_total
        self.bar = tqdm(
            desc=desc,
            total=bar_total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )

    def __call__(self, elapsed: Elapsed, batch: Optional[Batch] = None) -> None:
        if self.total is None or self.total.steps is not None:
            if self.prev_step is None:
                self.prev_step = elapsed.steps - 1
            self.bar.update(elapsed.steps - self.prev_step)
            self.prev_step = elapsed.steps
        elif self.total.samples is not None:
            if self.prev_samples is None:
                if batch is None:
                    raise ValueError("Batch is required for sample-based progress bar")
                self.prev_samples = elapsed.samples - get_batch_size(batch)
            self.bar.update(elapsed.samples - self.prev_samples)
            self.prev_samples = elapsed.samples
        elif self.total.time is not None:
            if self.prev_time is None:
                self.prev_time = elapsed.date_start
            self.bar.update(elapsed.date - self.prev_time)
            self.prev_time = elapsed.date
        else:
            raise ValueError("Invalid total")

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        self(loop_state.elapsed, loop_state.batch)
        return {}, loop_state.state


class keras_bar(LoopCallbackBase[S]):
    def __init__(
        self,
        total: Union[Period, int, None] = None,
        epoch=None,
        num_epochs=None,
        width=30,
        verbose=1,
        interval=0.05,
        stateful_metrics=None,
        always_stateful=False,
        unit_name="step",
    ):
        if isinstance(total, int):
            total = Period.create(steps=total)

        if total is not None:
            if total.steps is not None:
                bar_total = total.steps
                unit_name = "step"
            elif total.samples is not None:
                bar_total = total.samples
                unit_name = "sample"
            elif total.time is not None:
                bar_total = total.time
                unit_name = "s"
            elif total.date is not None:
                total = replace(total, time=total.date - datetime.now().timestamp())
                bar_total = total.time
                unit_name = "s"
            else:
                raise ValueError("Invalid total")
        else:
            bar_total = None
            unit_name = "it"

        self.total = total
        self.prev_step: Optional[int] = None
        self.prev_samples: Optional[int] = None
        self.prev_time: Optional[float] = None
        self.bar_total = bar_total
        self.bar = Kbar(
            bar_total,
            epoch=epoch,
            num_epochs=num_epochs,
            width=width,
            verbose=verbose,
            interval=interval,
            stateful_metrics=stateful_metrics,
            always_stateful=always_stateful,
            unit_name=unit_name,
        )

    def __call__(self, elapsed: Elapsed, logs: Logs) -> None:
        if self.total is None or self.total.steps is not None:
            current = elapsed.steps
        elif self.total.samples is not None:
            current = elapsed.samples
        elif self.total.time is not None:
            current = elapsed.time
        else:
            raise ValueError("Invalid total")

        metrics: Dict[str, Any] = {}
        if "stateful_metrics" in logs:
            stateful_metrics = logs["stateful_metrics"]
            self.bar.stateful_metrics.update(stateful_metrics.keys())
            metrics.update(stateful_metrics)

        if "metrics" in logs:
            metrics.update(logs["metrics"])

        if metrics:
            self.bar.update(
                current,
                values=[(k, v) for k, v in metrics.items() if is_scalar(v)],
            )

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        self(loop_state.elapsed, loop_state.logs)
        return Logs(), loop_state.state

    def on_train_batch_end(self, state, batch, elapsed, loop_state: LoopState[S]):
        return self.__loop_callback__(loop_state)


if importlib.util.find_spec("wandb") is not None:
    from wandb.wandb_run import Run

    class wandb_logger(LoopCallbackBase[S]):
        def __init__(self, run: Run):
            from wandb.wandb_run import Run

            self.run: Run = run

        def __call__(self, elapsed: Elapsed, logs: Logs) -> None:
            data = {}
            for collection, collection_logs in logs.items():
                for key, value in collection_logs.items():
                    if is_scalar(value):
                        if key in data:
                            key = f"{collection}.{key}"
                        data[key] = value

            if len(data) > 0:
                self.run.log(data, step=elapsed.steps)

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self(loop_state.elapsed, loop_state.logs)
            return Logs(), loop_state.state

else:
    wandb_logger = unavailable_dependency(
        "'wandb' package is not available, please install it to use the 'wandb_logger' callback"
    )


class NoOp(LoopCallbackBase[S]):
    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        return Logs(), loop_state.state


noop = NoOp()

# -------------------------------------------
# Adapters
# -------------------------------------------

if importlib.util.find_spec("clu") is not None:
    from clu.periodic_actions import PeriodicAction

    @dataclass(frozen=True)
    class PeriodicActionCallback(LoopCallbackBase[S]):
        action: PeriodicAction

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self.action(loop_state.elapsed.steps, t=loop_state.elapsed.date)
            return Logs(), loop_state.state

    @functools.partial(register_adapter, cls=PeriodicAction)
    def periodic_action_adapter(f: PeriodicAction):
        return PeriodicActionCallback(f)
